import sys
import ssl
import yaml
import logging
import asyncio
import argparse
import certstream
from .processor import Processor
from .constants import *
from .output.logoutputsink import LogOutputSink
from .output.fileoutputsink import FileOutputSink
from .output.weboutputsink import WebOutputSink

proc = None  # Populated within main()
config = {}  # Populated after YAML file parsing
output_sinks = []  # Populated after init_output_sinks()

logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)


def cert_callback(message, context):
    """CertStream main callback function"""
    domain = message["data"]["leaf_cert"]["all_domains"][0]

    # Skip whitelisted keywords
    for wl_keyword in config[ATTR_WHITELIST]:
        if wl_keyword in domain:
            return

    result = proc.process(domain)
    if result is not None:
        for sink in output_sinks:
            # Special case for WebOutputSink - the send_output() method is declared as async, 
            # in order to issue a non-blocking request
            if isinstance(sink, WebOutputSink): 
                asyncio.run(sink.send_output(domain, result))
            else:
                sink.send_output(domain, result)


def on_error(instance, exception=None):
    logging.error("[!] CertStream error: {}".format(exception))


def read_config(filename=DEFAULT_CONFIG):
    """Parses YAML file for configuration"""
    try:
        with open(filename, "r") as f:
            new_config = yaml.safe_load(f)
            new_config["keywords"] = set(new_config["keywords"])
            return new_config

    except Exception as e:
        logging.error("[!] Failed to open {}. {}".format(filename, e))
        sys.exit(1)


def validate_config(config):
    """Checks whether required attributes are set and inserts defaults where possible"""
    if ATTR_KEYWORDS not in config or len(config[ATTR_KEYWORDS]) == 0:
        logging.error("[!] Keywords not configured.".format(ATTR_KEYWORDS))
        sys.exit(1)

    # By default, whitelist is empty
    if ATTR_WHITELIST not in config:
        config[ATTR_WHITELIST] = []

    # If a certstream server URL is not set, use the one hosted by Calidog
    if ATTR_CERTSTREAM_URL not in config:
        config[ATTR_CERTSTREAM_URL] = DEFAULT_CERTSTREAM

    # If an output sink isn't configured, just enable console logging
    if ATTR_OUTPUT not in config:
        console_config = {ATTR_OUTPUT_CONSOLE: True}
        config[ATTR_OUTPUT] = console_config

    # If a webhook sink is specified, make sure it has at least a URL
    if ATTR_OUTPUT_WEBHOOK in config and ATTR_OUTPUT_WEBHOOK_URL not in config[ATTR_OUTPUT_WEBHOOK]:
        logging.error("[!] URL not specified for webhook sink.")
        sys.exit(1)

    # Default threshold is 1
    if ATTR_THRESHOLD not in config or int(config[ATTR_THRESHOLD]) <= 0:
        config[ATTR_THRESHOLD] = DEFAULT_THRESHOLD

    # Be default, verify secure connection with certstream-server
    if ATTR_DISABLE_TLS not in config:
        config[ATTR_DISABLE_TLS] = False

    return config


def init_output_sinks(config):
    """Creates output sinks according to configuration file"""
    sinks = []

    # Type 1: Console log output
    if ATTR_OUTPUT_CONSOLE in config[ATTR_OUTPUT]:
        sinks.append(LogOutputSink(logger))

    # Type 2: File output
    if ATTR_OUTPUT_FILE in config[ATTR_OUTPUT]:
        try:
            filename = config[ATTR_OUTPUT][ATTR_OUTPUT_FILE]
            sinks.append(FileOutputSink(filename))
        except RuntimeError as e:
            logger.error(e)
            sys.exit(1)

    # Type 3: Webhook output
    if ATTR_OUTPUT_WEBHOOK in config[ATTR_OUTPUT]:
        url = config[ATTR_OUTPUT][ATTR_OUTPUT_WEBHOOK][ATTR_OUTPUT_WEBHOOK_URL]
        body = config[ATTR_OUTPUT][ATTR_OUTPUT_WEBHOOK][ATTR_OUTPUT_WEBHOOK_BODY] if ATTR_OUTPUT_WEBHOOK_BODY in config[ATTR_OUTPUT][ATTR_OUTPUT_WEBHOOK] else None
        sinks.append(WebOutputSink(url, body))

    return sinks


def show_banner():
    banner = r'''
  _____  _     _     _     _____           _            
 |  __ \| |   (_)   | |   |  __ \         | |           
 | |__) | |__  _ ___| |__ | |__) |__ _  __| | __ _ _ __ 
 |  ___/| '_ \| / __| '_ \|  _  // _` |/ _` |/ _` | '__|
 | |    | | | | \__ \ | | | | \ \ (_| | (_| | (_| | |   
 |_|    |_| |_|_|___/_| |_|_|  \_\__,_|\__,_|\__,_|_|   
                                                        
                                                                                                          
(c) 2025 Narek Babajanyan
'''
    print(banner)


def main(arguments=()):
    """Program entry point. Reads config file and connects to the CertStream server"""
    show_banner()
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", help="configuration YAML file path", required=True)
    args = parser.parse_args()

    config_file = args.config if args.config else DEFAULT_CONFIG
    logger.info("[*] Retrieving configurations from {}".format(config_file))

    global config
    config = validate_config(read_config(config_file))

    global proc
    proc = Processor(config[ATTR_THRESHOLD], config[ATTR_KEYWORDS])

    global output_sinks
    output_sinks = init_output_sinks(config)

    logger.info(
        "[*] Begin monitoring for the following keywords: {}".format(
            config[ATTR_KEYWORDS]
        )
    )

    # Disable SSL/TLS verification, if configured.
    ssl_opt = {"cert_reqs":ssl.CERT_NONE} if config[ATTR_DISABLE_TLS] else None

    certstream.listen_for_events(
        cert_callback, on_error=on_error, url=config[ATTR_CERTSTREAM_URL], sslopt=ssl_opt
    )


if __name__ == "__main__":
    main(sys.argv[1:])



from .constants import RARE_KW_SCORE
import wordsegment


class Processor(object):
    threshold = 1
    keyword_scores = {}

    def __init__(self, threshold, keywords):
        wordsegment.load()
        self.threshold = threshold

        # The wordsegment package does not publicly expose the score() method,
        # but we need it in order to determine rare words.
        # This violation of private variable convention is a necessary evil.
        for keyword in keywords:
            self.keyword_scores[keyword] = wordsegment._segmenter.score(keyword)

    def process(self, domain):
        matches = set()

        # For rare keywords, just check if the domain contains them
        for keyword in self.keyword_scores:
            if self.keyword_scores[keyword] <= RARE_KW_SCORE and keyword in domain:
                matches.add(keyword)

        # ...otherwise, proceed with Wordsegment
        segments = wordsegment.segment(
            str(domain).lower().replace(".", " ").replace("-", " ")
        )
        
        matches = matches.union(set(self.keyword_scores.keys()).intersection(set(segments)))
        if len(matches) >= self.threshold:
            return matches
        else:
            return None



from .abstractoutputsink import AbstractOutputSink
from ..constants import *


class FileOutputSink(AbstractOutputSink):

    def __init__(self, log_filename):
        # Check if file can be written to during initialization
        try:
            open(log_filename, "a+").close()
            self.logfile = log_filename
        except OSError as e:
            raise RuntimeError(
                f"[!] Failed to open {log_filename} for logging! {e}"
            ) from e

    def send_output(self, domain, output_matches):
        try:
            with open(self.logfile, "a+") as f:
                f.write(
                    f"[+] {domain} likely contains the following keywords: {output_matches}\r\n"
                )
        except OSError as e:
            raise RuntimeError(
                f"[!] Failed to open {self.logfile} for logging! {e}"
            ) from e



from ..processor import Processor

# "Testarossa" is a known keyword, retrieved via Wordsegment
def test_process_common_keywords():
    threshold = 1
    keywords = ["test", "testarossa"]
    domain = "ferraritestarossa.it"
    proc = Processor(threshold, keywords)
    assert "testarossa" in proc.process(domain)

# "Testarosso" is not a known keyword and is not recognized by Wordsegment
def test_process_rare_keywords():
    threshold = 1
    keywords = ["test", "testarosso"]
    domain = "ferraritestarosso.it"
    proc = Processor(threshold, keywords)
    assert "testarosso" in proc.process(domain)


from setuptools import setup, find_packages

with open("README.md", "r") as mdf:
    long_description = mdf.read()

setup(
    name="phishradar",
    version="0.3",
    description="Phishing domain detection from Certificate Transparency logs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/narekb/PhishRadar",
    author="Narek Babajanyan",
    author_email="narek_babajanyan@outlook.com",
    license="Apache Software License (Apache 2.0)",
    project_urls={
        "Source": "https://github.com/narekb/PhishRadar"
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Information Technology",
        "Topic :: Security"
    ],
    python_requires=">=3.13",
    install_requires=["certstream", "PyYAML", "wordsegment", "aiohttp"],
    packages=find_packages(),
    include_package_data=True,
    entry_points = {
        "console_scripts": [
            "phishradar = phishradar:main"
        ] 
    }
)


DEFAULT_CONFIG = "config.yaml"
DEFAULT_CERTSTREAM = "wss://certstream.calidog.io/"
DEFAULT_THRESHOLD = 1

# Configuration attributes
ATTR_CERTSTREAM_URL = "certstream_url"
ATTR_KEYWORDS = "keywords"
ATTR_WHITELIST = "whitelist"
ATTR_OUTPUT = "output"
ATTR_OUTPUT_CONSOLE = "console"
ATTR_OUTPUT_FILE = "file"
ATTR_OUTPUT_WEBHOOK = "webhook"
ATTR_OUTPUT_WEBHOOK_URL = "url"
ATTR_OUTPUT_WEBHOOK_BODY = "body"
ATTR_THRESHOLD = "threshold"
ATTR_DISABLE_TLS = "disable_tls"

# This value was discovered experimentally by comparing scores of common and rare words (e.g. brand names)
RARE_KW_SCORE = 0.0000000005



import logging
from .abstractoutputsink import AbstractOutputSink


class LogOutputSink(AbstractOutputSink):
    def __init__(self, logger):
        self.logger = logger

    def send_output(self, domain, output_matches):
        self.logger.info(
            f"[+] {domain} likely contains the following keywords: {output_matches}"
        )



from .abstractoutputsink import AbstractOutputSink
from string import Template
import aiohttp

class WebOutputSink(AbstractOutputSink):
    def __init__(self, url, body):
        self.url = url
        self.body = body

    def bulleted_list(self, matches):
        list_str = ""
        for match in matches:
            list_str += ("\r\n• " + match)
        return list_str

    async def send_output(self, domain, output_matches):
        # Defang the domain so that it's not clickable in other environments
        domain = domain.replace(".", "[.]")

        # Provide multiple template formats for matches - comma-separated and bulleted list
        matches_comma = ", ".join(output_matches)
        matches_list = self.bulleted_list(output_matches)

        # Insert specified templates into the URL and request body

        url_template = Template(self.url)
        body_template = Template(self.body)

        url = url_template.substitute(domain=domain, matches_comma=matches_comma, matches_list=matches_list)
        body = body_template.substitute(domain=domain, matches_comma=matches_comma, matches_list=matches_list)
        
        async with aiohttp.ClientSession() as session:
            async with session.post(url=url, data=body, headers={"Content-Type": "application/json"}) as response:
                await response.release()